/* 
 * lib.c
 * simple stack  file server library
 *
 */

#include <stackfs.h>
#include "node.h"

#define DEBUG 1


static Node *rootnode;
static char *realpath;

static Stackfs *sfs;


int
dprint(char *fmt, ...)
{
	int n;
	va_list args;


	if(! DEBUG)
		return -1;
	va_start(args, fmt);
	n = vfprint(2, fmt, args);
	va_end(args);
	return n;
}

/* initialize function */
void
init(void)
{	
	setpath(sfs->realpath);
	rootnode = newnode(nil, "");
	return;
}


static void
_fsattach(Req *r)
{
	char *spec;
	
	if(sfs->attach){
		sfs->attach(r);
		return;
	}

	/* this fs provides single tree */
	dprint("fsattach()\n");
	spec = r->ifcall.aname;
	if(spec && spec[0]){
		respond(r, "invalid attach specifier");
		return;
	}

	r->fid->aux = rootnode;
	r->fid->qid = ((Node*)r->fid->aux)->d->qid;
	r->ofcall.qid = r->fid->qid;
	respond(r, nil);
}


static char*
_fswalk1(Fid *fid, char *name, Qid *qid)
{
	int i, fd;
	long n;
	Dir *db;
	Node *node, *np;

	if(sfs->walk1){
		return sfs->walk1(fid, name, qid);
	}

	dprint("fswalk()\n");
	node  = fid->aux;

	if(strcmp(name, "..")==0){
		if(node->parent){
			fid->aux = node->parent;
			fid->qid = node->parent->d->qid;
			*qid = fid->qid;
		}else{
			*qid = fid->qid;
		}

		return nil;
	}


	/* read dir and create node */

	fd = open(node->realname, OREAD);
	if(fd == -1)
		return "can't open Directory";

	n = dirreadall(fd, &db);
	close(fd);

	for(i=0;i<n;i++){
		if(strcmp((db+i)->name, name) == 0){
			np = newnode(node, name);
			fid->aux = np;
			fid->qid = np->d->qid;
			*qid=fid->qid;
			return nil;
		}
	}

	return "not found";	
}

static char*
_fsclone(Fid  *ofid, Fid *fid)
{
	Node *node, *onode;

	if(sfs->clone){
		return sfs->clone(ofid, fid);
	}

	dprint("fsclone()\n");
	onode = (Node*)ofid->aux;
	node = emalloc(sizeof(Node));
	*node = *onode;
	node->d = reallocdir(onode->d,0);
	fid->aux = node;
	return nil;
}


/* fill stat information */
static void
fillstat(Dir *dir, Dir *d)
{
	if(sfs->fillstat){
		sfs->fillstat(dir, d);
		return;
	}

	dprint("fillstat()\n");
	dprint("\td->name: %s\n",d->name);

	dprint("fillstat -> reallocdir()\n");

	*dir = *d;
	dir->name = estrdup(d->name);
	dir->uid = estrdup(d->uid);
	dir->gid = estrdup(d->gid);
	dir->muid = estrdup(d->muid);

	return;
}

static void
wfillstat(Dir *dir, Node *np)
{
	if(sfs->wfillstat){
		sfs->wfillstat(dir, np);
		return;
	}

	dprint("wfillstat()\n");
	if(dirwstat(np->realname, dir) <0){
		dprint("wstat error\n");
	}

}

static int
fsdirgen(int n, Dir *dir, void *p)
{
	int fd;
	long m;
	Dir *db;
	Node *node;
	node = (Node*)p;

	dprint("fsdirgen()\n");

	fd = open(node->realname, OREAD);
	if(fd == -1)
		return -1;

	m = dirreadall(fd, &db);

	if(n < m){
		fillstat(dir, db+n);
		close(fd);
		return 0;
	}

	close(fd);
	return -1;
}

static void
_fsstat(Req *r)
{
	dprint("fsstat()\n");

	if(r->fid->aux){
		fillstat(&(r->d), ((Node*)r->fid->aux)->d);
		respond(r, nil);
		return;
	}

	respond(r, "something error in fsstat()");
}

static void
_fswstat(Req *r)
{
	Node *np;
	Dir *d, *dir;


	dir  = &(r->d);

	dprint("fswstat()\n");
	if(r->fid->aux){
		np = (Node *)r->fid->aux;
		if(dirwstat(np->realname, dir) < 0){
			respond(r, "wstat error");
			return;
		}
		d = np->d;

		*d = *dir;
		d->name = estrdup(dir->name);
		d->uid = estrdup(dir->uid);
		d->gid = estrdup(dir->gid);
		d->muid = estrdup(dir->muid);
		respond(r, nil);
		return;
	}
	respond(r, "somthing error in fswstat()");
	
}



/* open the file */
static void
_fsopen(Req *r)
{
	int fd;
	Node *node;

	if(sfs->open){
		sfs->open(r);
		return;
	}


	node = ((Node*)r->fid->aux);
	dprint("fsopen()\n");

	if( node->open ){
		close( node->fd );
		node->open = 0;
	}

	fd = open(node->realname, r->ifcall.mode);
	if(fd == -1){
		respond(r, "permission denied");
		return;
	}
	
	node->fd = fd;
	node->open = 1;

	respond(r, nil);
}
			
/* destroy fid */
static void
_fsdestroyfid(Fid *fid)
{
	Node *node;

	if(sfs->destroyfid){
		sfs->destroyfid(fid);
		return;
	}

	dprint("fsdestroyfid()\n");
	node = (Node*)fid->aux;

	if(node==nil){
		return;
	}

	if(node->open){
		close(node->fd);
		dprint("\tclose fd\n");
	}

	free(node->d);
	free(node);
}

/* read file */
static void
_fsread(Req *r)
{
	Node *node;
	char buf[8*1024];
	ulong len;
	int fd;


	if(sfs->read){
		sfs->read(r);
		return;
	}

	node = (Node*)r->fid->aux;
	dprint("fsread(%s)\n",node->realname);
	
	if(node->d->qid.type == QTDIR){
		dirread9p(r, fsdirgen, r->fid->aux);
	}else{
		fd = node->fd;
		if(len = pread(fd, buf, sizeof(buf),r->ifcall.offset) ){

			if(sfs->readfilter)
				sfs->readfilter(buf);

			readbuf(r, buf, len);
		}
	}

	respond(r,nil);
}




static void
_fswrite(Req *r)
{

	int fd;
	long len;
	Node * node;
	char *buf;

	if(sfs->write){
		sfs->write(r);
		return;
	}

	node = (Node *)r->fid->aux;

	dprint("fswrite()\n");
	/* do nothing, just respond error */
	if(r->fid->qid.type&QTDIR)
		respond(r, "directories are not writable");


	fd = node->fd;
	buf = r->ifcall.data;
	
	if(sfs->writefilter)
		sfs->writefilter(buf);

	
	if( (len = pwrite(fd, r->ifcall.data, r->ifcall.count, r->ifcall.offset)) >=0){
		r->ofcall.count = len;
		respond(r, nil);
	}else{
		respond(r, "something error in write");
	}
}

void
_fscreate(Req *r)
{
	char *name;
	ulong perm;
	int mode;
	Node *np;
	int fd;

	if(sfs->create){
		sfs->create(r);
		return;
	}


	dprint("fscreate()\n");

	if( ((Node*)r->fid->aux)->d->qid.type&QTDIR == 0){
		respond(r, "not a directory");
		return;
	}

	
	name = emalloc(strlen(((Node*)r->fid->aux)->realname)+strlen(r->ifcall.name)+2);
	sprint(name, "%s/%s",((Node*)r->fid->aux)->realname, r->ifcall.name);
	perm = r->ifcall.perm;
	mode = r->ifcall.mode;
	
	if(fd = create(name, mode, perm) < 0){
		dprint("\tcreate failed\n");
		respond(r, "permission denied");
		return;
	}

	np = newnode(r->fid->aux, r->ifcall.name);
	np->fd = fd;
	np->open = 1;
	r->fid->aux = np;
	r->fid->qid = np->d->qid;
	r->ofcall.qid = r->fid->qid;

	respond(r, nil);
	
}


static Srv srvfs = {
	.attach	= _fsattach,
	.walk1	= _fswalk1,
	.stat		= _fsstat,
	.wstat	= _fswstat,
	.clone	= _fsclone,
	.destroyfid = _fsdestroyfid,
	.open	= _fsopen,
	.read		= _fsread,
	.write	= _fswrite,
	.create	= _fscreate
	
};



void
poststackfs(Stackfs *srv, char *srvname,  char *mtpt, int flag)
{
	sfs = srv;
	init();
	postmountsrv(&srvfs, srvname, mtpt, flag);
}